/*
GNU General Public License

BMPCodec
Copyright (C) 2000  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.codec;

import java.awt.AWTException;
import java.awt.Image;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.awt.image.PixelGrabber;

/** BMPCodec
  *
  * @author Howard Kistler
  * @version 0.1
  *
  * VERSION HISTORY
  * 0.1 (10/26/2000) - initial creation (10/24/2000)
  *
  * KNOWN BUGS & LIMITATIONS
  * 1) Only makes 24-bit BMP files
  * 2) Seems to shift the bits after the first scanline, resulting in a one-pixel wrap effect (padding problem?)
  *
  * ACKNOWLEDGEMENT
  * Originally derived from code example by Jean-Pierre Dub
  */

public class BMPCodec
{

//  Constants ---------------------------------------------------------------------------->

	// Header Array Sizes
	private final static byte FileHeaderSize             = (byte)14; // FileHeader is always 14 bytes in size for BMP format
	private final static byte InfoHeaderSize             = (byte)40; // InfoHeader is always 40 bytes in size for BMP/ICO/CUR format

	// File Header Fields
	private final static byte TypeByte1                  = 'B'; // Bitmaps are always type "BM"
	private final static byte TypeByte2                  = 'M';
	private final static byte ReservedByte1              = (byte)0;
	private final static byte ReservedByte2              = (byte)0;
	private final static byte ReservedByte3              = (byte)0;
	private final static byte ReservedByte4              = (byte)0;
	private final static byte MapOffsetByte1             = (byte)((FileHeaderSize + InfoHeaderSize)         & 0xFF);
	private final static byte MapOffsetByte2             = (byte)(((FileHeaderSize + InfoHeaderSize) >> 8)  & 0xFF);
	private final static byte MapOffsetByte3             = (byte)(((FileHeaderSize + InfoHeaderSize) >> 16) & 0xFF);
	private final static byte MapOffsetByte4             = (byte)(((FileHeaderSize + InfoHeaderSize) >> 24) & 0xFF);

	// Bitmap Info Header Fields
	private final static byte InfoHeaderSizeByte1        = (byte)(InfoHeaderSize         & 0xFF);
	private final static byte InfoHeaderSizeByte2        = (byte)((InfoHeaderSize >> 8)  & 0xFF);
	private final static byte InfoHeaderSizeByte3        = (byte)((InfoHeaderSize >> 16) & 0xFF);
	private final static byte InfoHeaderSizeByte4        = (byte)((InfoHeaderSize >> 24) & 0xFF);
	private final static byte BitmapPlanesByte1          = (byte)1; // Bitmaps always have 1 bitplane
	private final static byte BitmapPlanesByte2          = (byte)0;
	private final static byte BitmapBitCountByte1        = (byte)24; // This codec creates 24-bit BMP files
	private final static byte BitmapBitCountByte2        = (byte)0;
	private final static byte BitmapCompressionByte1     = (byte)0;
	private final static byte BitmapCompressionByte2     = (byte)0;
	private final static byte BitmapCompressionByte3     = (byte)0;
	private final static byte BitmapCompressionByte4     = (byte)0;
	private final static byte BitmapXPixelsPerMByte1     = (byte)0;
	private final static byte BitmapXPixelsPerMByte2     = (byte)0;
	private final static byte BitmapXPixelsPerMByte3     = (byte)0;
	private final static byte BitmapXPixelsPerMByte4     = (byte)0;
	private final static byte BitmapYPixelsPerMByte1     = (byte)0;
	private final static byte BitmapYPixelsPerMByte2     = (byte)0;
	private final static byte BitmapYPixelsPerMByte3     = (byte)0;
	private final static byte BitmapYPixelsPerMByte4     = (byte)0;
	private final static byte BitmapColorsUsedByte1      = (byte)0;
	private final static byte BitmapColorsUsedByte2      = (byte)0;
	private final static byte BitmapColorsUsedByte3      = (byte)0;
	private final static byte BitmapColorsUsedByte4      = (byte)0;
	private final static byte BitmapColorsImportantByte1 = (byte)0;
	private final static byte BitmapColorsImportantByte2 = (byte)0;
	private final static byte BitmapColorsImportantByte3 = (byte)0;
	private final static byte BitmapColorsImportantByte4 = (byte)0;

//  Variables ---------------------------------------------------------------------------->

	private static int IntBitmapWidth;    // Width of image in pixels
	private static int IntBitmapHeight;   // Height of image in pixels
	private static int IntBitmapSize;     // Size of bitmap data
	private static int IntBitmapFileSize; // Size of entire bitmap file, including headers
	private static int[] PixelMap;        // Pixel array of image data

//  Constructor -------------------------------------------------------------------------->

	public BMPCodec()
	{
	}

//  Encode Method ------------------------------------------------------------------------>

	public static void encode(Image image, File BMPfile) throws FileNotFoundException, IOException, AWTException
	{
		convertToFormat(image);
		writeImageToFile(BMPfile);
	}

	public static void encode(Image image, String BMPfilename) throws FileNotFoundException, IOException, AWTException
	{
		encode(image, new File(BMPfilename));
	}

//  Image Convert Method ---------------------------------------------------------------->

	public static void convertToFormat(Image image) throws AWTException
	{
		IntBitmapWidth  = image.getWidth(null);
		IntBitmapHeight = image.getHeight(null);

		PixelMap = new int[IntBitmapWidth * IntBitmapHeight];

		PixelGrabber pg = new PixelGrabber(image, 0, 0, IntBitmapWidth, IntBitmapHeight, PixelMap, 0, IntBitmapWidth);
		try
		{
			pg.grabPixels();
		}
		catch(InterruptedException ie)
		{
			ie.printStackTrace(System.out);
		}

		int BitPadding = (4 - ((IntBitmapWidth * 3) % 4));
		if(BitPadding == 4) { BitPadding = 0; }
		BitPadding = BitPadding * IntBitmapHeight;

		IntBitmapSize = ((IntBitmapWidth * IntBitmapHeight) * 3) + BitPadding;

		IntBitmapFileSize = FileHeaderSize + InfoHeaderSize + IntBitmapSize;
	}

//  File Write Method ------------------------------------------------------------------->

	protected static void writeImageToFile(File BMPfile) throws IOException
	{
		FileOutputStream fos = new FileOutputStream(BMPfile);

		byte FileSizeByte1 = (byte)(IntBitmapFileSize         & 0xFF);
		byte FileSizeByte2 = (byte)((IntBitmapFileSize >> 8)  & 0xFF);
		byte FileSizeByte3 = (byte)((IntBitmapFileSize >> 16) & 0xFF);
		byte FileSizeByte4 = (byte)((IntBitmapFileSize >> 24) & 0xFF);

		byte[] BitmapFileHeader = new byte[]
		{
			TypeByte1,       TypeByte2,
			FileSizeByte1,   FileSizeByte2,   FileSizeByte3,   FileSizeByte4,
			ReservedByte1,   ReservedByte2,   ReservedByte3,   ReservedByte4,
			MapOffsetByte1,  MapOffsetByte2,  MapOffsetByte3,  MapOffsetByte4
		};
		fos.write(BitmapFileHeader);

		byte BitmapWidthByte1      = (byte)(IntBitmapWidth          & 0xFF);
		byte BitmapWidthByte2      = (byte)((IntBitmapWidth >> 8)   & 0xFF);
		byte BitmapWidthByte3      = (byte)((IntBitmapWidth >> 16)  & 0xFF);
		byte BitmapWidthByte4      = (byte)((IntBitmapWidth >> 24)  & 0xFF);
		byte BitmapHeightByte1     = (byte)(IntBitmapHeight         & 0xFF);
		byte BitmapHeightByte2     = (byte)((IntBitmapHeight >> 8)  & 0xFF);
		byte BitmapHeightByte3     = (byte)((IntBitmapHeight >> 16) & 0xFF);
		byte BitmapHeightByte4     = (byte)((IntBitmapHeight >> 24) & 0xFF);
		byte BitmapImageSizeByte1  = (byte)(IntBitmapSize           & 0xFF);
		byte BitmapImageSizeByte2  = (byte)((IntBitmapSize >> 8)    & 0xFF);
		byte BitmapImageSizeByte3  = (byte)((IntBitmapSize >> 16)   & 0xFF);
		byte BitmapImageSizeByte4  = (byte)((IntBitmapSize >> 24)   & 0xFF);

		byte[] BitmapBodyHeader = new byte[] {
			InfoHeaderSizeByte1,        InfoHeaderSizeByte2,        InfoHeaderSizeByte3,        InfoHeaderSizeByte4,
			BitmapWidthByte1,           BitmapWidthByte2,           BitmapWidthByte3,           BitmapWidthByte4,
			BitmapHeightByte1,          BitmapHeightByte2,          BitmapHeightByte3,          BitmapHeightByte4,
			BitmapPlanesByte1,          BitmapPlanesByte2,
			BitmapBitCountByte1,        BitmapBitCountByte2,
			BitmapCompressionByte1,     BitmapCompressionByte2,     BitmapCompressionByte3,     BitmapCompressionByte4,
			BitmapImageSizeByte1,       BitmapImageSizeByte2,       BitmapImageSizeByte3,       BitmapImageSizeByte4,
			BitmapXPixelsPerMByte1,     BitmapXPixelsPerMByte2,     BitmapXPixelsPerMByte3,     BitmapXPixelsPerMByte4,
			BitmapYPixelsPerMByte1,     BitmapYPixelsPerMByte2,     BitmapYPixelsPerMByte3,     BitmapYPixelsPerMByte4,
			BitmapColorsUsedByte1,      BitmapColorsUsedByte2,      BitmapColorsUsedByte3,      BitmapColorsUsedByte4,
			BitmapColorsImportantByte1, BitmapColorsImportantByte2, BitmapColorsImportantByte3, BitmapColorsImportantByte4
		};
		fos.write(BitmapBodyHeader);

		int bodySize = IntBitmapWidth * IntBitmapHeight;
		int bitPad = 4 - ((IntBitmapWidth * 3) % 4);
		if(bitPad == 4) { bitPad = 0; }
		int countRow = 1;
		int indexRow = bodySize - IntBitmapWidth;
		int indexLastRow = indexRow;
		byte[] rgbArray = new byte[3];

		for(int i = 0; i < bodySize; i++)
		{
			int colorValue = PixelMap[indexRow];
			rgbArray[0] = (byte)(colorValue         & 0xFF); // red
			rgbArray[1] = (byte)((colorValue >> 8)  & 0xFF); // green
			rgbArray[2] = (byte)((colorValue >> 16) & 0xFF); // blue
			fos.write(rgbArray);
			if(countRow == IntBitmapWidth)
			{
				// pad row to 4 bits requirement
				for(int p = 0; p < bitPad; p++)
				{
					fos.write(0x00);
				}
				countRow = 1;
				indexRow = indexLastRow - IntBitmapWidth;
				indexLastRow = indexRow;
			}
			else
			{
				countRow++;
			}
			indexRow++;
		}

		fos.flush();
		fos.close();
	}

}
