/*
GNU General Public License

GIFCodec
Copyright (C) 2000  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.codec;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.awt.AWTException;
import java.awt.Image;
import java.awt.image.PixelGrabber;

/** GIFCodec
  *
  * @author Howard Kistler
  * @version 0.1
  *
  * VERSION HISTORY
  * 0.1 (10/30/2000) - initial creation (10/24/2000)
  *
  * ACKNOWLEDGEMENT
  * Originally derived from code example by Adam Doppelt
  */

public class GIFCodec
{

//  Constants ---------------------------------------------------------------------------->

	// Convenience Constants
	private final static int MAXCOLORS = 256;

	// File Header Fields
	private final static byte TypeByte1 = 'G'; // Writes GIFs of type GIF89a
	private final static byte TypeByte2 = 'I';
	private final static byte TypeByte3 = 'F';
	private final static byte TypeByte4 = '8';
	private final static byte TypeByte5 = '9';
	private final static byte TypeByte6 = 'a';

//  Variables ---------------------------------------------------------------------------->

	private static int    IntBitmapWidth;
	private static int    IntBitmapHeight;
	private static int    ColorsUsed;
	private static byte[] PixelMap;
	private static byte[] ColorMap;

//  Constructor ----------------------------------------------------------------------------->

	public GIFCodec()
	{
	}

//  Encode Method ------------------------------------------------------------------------>

	public static void encode(Image image, File GIFfile) throws FileNotFoundException, IOException, AWTException
	{
		convertToFormat(image);
		writeImageToFile(GIFfile);
	}

	public static void encode(Image image, String GIFfilename) throws FileNotFoundException, IOException, AWTException
	{
		encode(image, new File(GIFfilename));
	}

//  Image Convert Method ---------------------------------------------------------------->

	protected static void convertToFormat(Image image) throws AWTException
	{
		IntBitmapWidth  = image.getWidth(null);
		IntBitmapHeight = image.getHeight(null);

		int[] BitMap = new int[IntBitmapWidth * IntBitmapHeight];

		PixelGrabber pg = new PixelGrabber(image, 0, 0, IntBitmapWidth, IntBitmapHeight, BitMap, 0, IntBitmapWidth);
		try
		{
			pg.grabPixels();
		}
		catch(InterruptedException ie)
		{
			ie.printStackTrace(System.out);
		}

		byte[][] rVals = new byte[IntBitmapWidth][IntBitmapHeight];
		byte[][] gVals = new byte[IntBitmapWidth][IntBitmapHeight];
		byte[][] bVals = new byte[IntBitmapWidth][IntBitmapHeight];

		for(int y = 0; y < IntBitmapHeight; y++)
		{
			for(int x = 0; x < IntBitmapWidth; x++)
			{
				rVals[x][y] = (byte)((BitMap[(y * IntBitmapWidth) + x] >> 16) & 0xFF);
				gVals[x][y] = (byte)((BitMap[(y * IntBitmapWidth) + x] >> 8)  & 0xFF);
				bVals[x][y] = (byte)( BitMap[(y * IntBitmapWidth) + x]        & 0xFF);
			}
		}
		buildColorMap(rVals, gVals, bVals);
	}

//  File Write Method ------------------------------------------------------------------->

	protected static void writeImageToFile(File GIFfile) throws IOException
	{
		FileOutputStream fos = new FileOutputStream(GIFfile);

		byte BitmapWidthByte1  = (byte)( IntBitmapWidth        & 0xFF);
		byte BitmapWidthByte2  = (byte)((IntBitmapWidth >> 8)  & 0xFF);
		byte BitmapHeightByte1 = (byte)( IntBitmapHeight       & 0xFF);
		byte BitmapHeightByte2 = (byte)((IntBitmapHeight >> 8) & 0xFF);

		byte MaskingByte = (byte)(-16);
		byte colorMask   = (byte)(calcNeededBits(ColorsUsed) - 1);
		MaskingByte      = (byte)(MaskingByte | (colorMask & 7));

		byte BackgroundColorIndexByte = (byte)0;

		byte AspectRatioByte = (byte)0;

		/* Build and write the file header */
		byte[] BitmapFileHeader = new byte[]
		{
			TypeByte1,         TypeByte2,         TypeByte3, TypeByte4, TypeByte5, TypeByte6,
			BitmapWidthByte1,  BitmapWidthByte2,
			BitmapHeightByte1, BitmapHeightByte2,
			MaskingByte,
			BackgroundColorIndexByte,
			AspectRatioByte
		};
		fos.write(BitmapFileHeader);

		/* Write the color map of color values and color indices */
		fos.write(ColorMap, 0, ColorMap.length);

		/* Build and write the image header */
		byte DataDelimiterByte  = ',';
		byte HorizontalPosByte1 = (byte)0;
		byte HorizontalPosByte2 = (byte)0;
		byte VerticalPosByte1   = (byte)0;
		byte VerticalPosByte2   = (byte)0;
		byte ImageMaskingByte   = (byte)0;

		byte[] BitmapImageHeader = new byte[]
		{
			DataDelimiterByte,
			HorizontalPosByte1, HorizontalPosByte2,
			VerticalPosByte1,   VerticalPosByte2,
			BitmapWidthByte1,   BitmapWidthByte2,
			BitmapHeightByte1,  BitmapHeightByte2,
			ImageMaskingByte
		};
		fos.write(BitmapImageHeader);

		/* Write the body size (based on number of colors in use) */
		byte bodySize = calcNeededBits(ColorsUsed);
		if(bodySize == 1) { bodySize++; }
		fos.write(bodySize);

		/* Compress the body into GIF LZW format */
///////////////////////<
		LZWCompressor.LZWCompress(fos, bodySize, PixelMap);
///////////////////////>

		/* Build and write the image footer */
		DataDelimiterByte  = ';';
		HorizontalPosByte1 = (byte)0;
		HorizontalPosByte2 = (byte)0;
		VerticalPosByte1   = (byte)0;
		VerticalPosByte2   = (byte)0;
		ImageMaskingByte   = (byte)0;

		byte[] BitmapImageFooter = new byte[]
		{
			DataDelimiterByte,
			HorizontalPosByte1, HorizontalPosByte2,
			VerticalPosByte1,   VerticalPosByte2,
			BitmapWidthByte1,   BitmapWidthByte2,
			BitmapHeightByte1,  BitmapHeightByte2,
			ImageMaskingByte
		};
		fos.write(BitmapImageFooter);

		fos.flush();
		fos.close();
	}

//  Internal Class Methods --------------------------------------------------------------->

	private static void buildColorMap(byte[][] r, byte[][] g, byte[][] b) throws AWTException
	{
		PixelMap = new byte[IntBitmapWidth * IntBitmapHeight]; // convert to 2-dim array in rewrite
		ColorMap = new byte[MAXCOLORS * 3]; // convert to 2-dim array in rewrite
		int colorIndex = 0;
		for(int y = 0; y < IntBitmapHeight; y++)
		{
			for(int x = 0; x < IntBitmapWidth; x++)
			{
				int lookupColor;
				for(lookupColor = 0; lookupColor < colorIndex; lookupColor++)
				{
					if(
						(ColorMap[(lookupColor * 3)]     == r[x][y]) &&
						(ColorMap[(lookupColor * 3) + 1] == g[x][y]) &&
						(ColorMap[(lookupColor * 3) + 2] == b[x][y])
					)
					{
						break;
					}
				}
				if(lookupColor >= MAXCOLORS)
				{
					throw new AWTException("Image contains more than " + MAXCOLORS + " colors");
				}
				PixelMap[(y * IntBitmapWidth) + x] = (byte)(lookupColor);
				// add to ColorMap if at the "new" position
				if(lookupColor == colorIndex)
				{
					ColorMap[(lookupColor * 3)]     = r[x][y];
					ColorMap[(lookupColor * 3) + 1] = g[x][y];
					ColorMap[(lookupColor * 3) + 2] = b[x][y];
					colorIndex++;
				}
			}
		}
		// Trim array to just the colors needed
		ColorsUsed = 1 << calcNeededBits(colorIndex);
		byte[] tempColorMap = new byte[ColorsUsed * 3]; // another candidate for 2-dim'ing
		System.arraycopy(ColorMap, 0, tempColorMap, 0, ColorsUsed * 3);
		ColorMap = tempColorMap;
	}

//  Convenience Methods --------------------------------------------------------------------->

	private static byte calcNeededBits(int count)
	{
		byte calcByte = 1;
		if(count-- == 0)
		{
			return (byte)0;
		}
		while((count >>= 1) != 0)
		{
			calcByte++;
		}
		return calcByte;
	}    

}

////////////////////////
//  Ancillary Classes ----------------------------------------------------------------------->

class BitFile
{
	OutputStream output_;
	byte buffer_[];
	int index_, bitsLeft_;

	public BitFile(OutputStream output)
	{
		output_ = output;
		buffer_ = new byte[256];
		index_ = 0;
		bitsLeft_ = 8;
	}

	public void flush() throws IOException
	{
		int numBytes = index_ + (bitsLeft_ == 8 ? 0 : 1);
		if(numBytes > 0)
		{
			output_.write(numBytes);
			output_.write(buffer_, 0, numBytes);
			buffer_[0] = 0;
			index_ = 0;
			bitsLeft_ = 8;
		}
	}

	public void writeBits(int bits, int numbits) throws IOException
	{
		int bitsWritten = 0;
		int numBytes = 255;
		do {
			if((index_ == 254 && bitsLeft_ == 0) || index_ > 254)
			{
				output_.write(numBytes);
				output_.write(buffer_, 0, numBytes);
				buffer_[0] = 0;
				index_ = 0;
				bitsLeft_ = 8;
			}

			if(numbits <= bitsLeft_)
			{
				buffer_[index_] |= (bits & ((1 << numbits) - 1)) << (8 - bitsLeft_);
				bitsWritten += numbits;
				bitsLeft_ -= numbits;
				numbits = 0;
			}
			else
			{
				buffer_[index_] |= (bits & ((1 << bitsLeft_) - 1)) << (8 - bitsLeft_);
				bitsWritten += bitsLeft_;
				bits >>= bitsLeft_;
				numbits -= bitsLeft_;
				buffer_[++index_] = 0;
				bitsLeft_ = 8;
			}
		} while(numbits != 0);
	}
}

class LZWStringTable
{
	private final static int   RES_CODES  = 2;
	private final static short HASH_FREE  = (short)0xFFFF;
	private final static short NEXT_FIRST = (short)0xFFFF;
	private final static int   MAXBITS    = 12;
	private final static int   MAXSTR     = (1 << MAXBITS);
	private final static short HASHSIZE   = 9973;
	private final static short HASHSTEP   = 2039;

	byte  strChar[];
	short strNext[];
	short strHash[];
	short numStrings;

	public LZWStringTable()
	{
		strChar = new byte[MAXSTR];
		strNext = new short[MAXSTR];
		strHash = new short[HASHSIZE];    
	}

	public int addCharString(short index, byte b)
	{
		int hashIndex;

		if(numStrings >= MAXSTR)
		{
			return 0xFFFF;
		}

		hashIndex = generateHash(index, b);
		while(strHash[hashIndex] != HASH_FREE)
		{
			hashIndex = (hashIndex + HASHSTEP) % HASHSIZE;
		}

		strHash[hashIndex]  = numStrings;
		strChar[numStrings] = b;
		strNext[numStrings] = ((index != HASH_FREE) ? index : NEXT_FIRST);

		return numStrings++;
	}

	public short findCharString(short index, byte b)
	{
		int hashIndex, nxtidx;

		if(index == HASH_FREE)
		{
			return b;
		}

		hashIndex = generateHash(index, b);
		while((nxtidx = strHash[hashIndex]) != HASH_FREE)
		{
			if(strNext[nxtidx] == index && strChar[nxtidx] == b)
			{
				return (short)nxtidx;
			}
			hashIndex = (hashIndex + HASHSTEP) % HASHSIZE;
		}

		return (short)0xFFFF;
	}

	public void clearTable(int codesize)
	{
		numStrings = 0;

		for(int q = 0; q < HASHSIZE; q++)
		{
			strHash[q] = HASH_FREE;
		}

		int w = (1 << codesize) + RES_CODES;
		for(int q = 0; q < w; q++)
		{
			addCharString((short)0xFFFF, (byte)q);
		}
	}

	static public int generateHash(short index, byte lastbyte)
	{
		return ((int)((short)(lastbyte << 8) ^ index) & 0xFFFF) % HASHSIZE;
	}
}

class LZWCompressor
{

	public static void LZWCompress(OutputStream output, int codesize, byte toCompress[]) throws IOException
	{
		byte c;
		short index;
		int clearcode, endofinfo, numbits, limit, errcode;
		short prefix = (short)0xFFFF;

		BitFile bitFile = new BitFile(output);
		LZWStringTable strings = new LZWStringTable();

		clearcode = 1 << codesize;
		endofinfo = clearcode + 1;

		numbits = codesize + 1;
		limit = (1 << numbits) - 1;

		strings.clearTable(codesize);
		bitFile.writeBits(clearcode, numbits);

		for(int i = 0; i < toCompress.length; i++)
		{
			c = toCompress[i];
			if((index = strings.findCharString(prefix, c)) != -1)
			{
				prefix = index;
			}
			else
			{
				bitFile.writeBits(prefix, numbits);
				if(strings.addCharString(prefix, c) > limit)
				{
					if(++numbits > 12)
					{
						bitFile.writeBits(clearcode, numbits - 1);
						strings.clearTable(codesize);
						numbits = codesize + 1;
					}
					limit = (1 << numbits) - 1;
				}
				prefix = (short)((short)c & 0xFF);
			}
		}

		if(prefix != -1)
		{
			bitFile.writeBits(prefix, numbits);
		}

		bitFile.writeBits(endofinfo, numbits);
		bitFile.flush();
	}
}
