/*
GNU General Public License

ICOCodec
Copyright (C) 2000  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.codec;

import java.awt.AWTException;
import java.awt.Image;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

/** ICOCodec
  *
  * @author Howard Kistler
  * @version 0.2
  *
  * VERSION HISTORY
  * 0.1 (08/29/2000) - initial creation (08/29/2000)
  * 0.2 (10/24/2000) - conversion to codec
  *
  * CURRENT LIMITATIONS
  * 1) BMP must be of size 16x16, 32x32, or 64x64
  * 2) BMP color depth must not be greater than 8-bit (256 colors)
  * 3) The alphamask (ANDBitMap) defaults to "all opaque"
  */

public class ICOCodec
{

//  Constants ---------------------------------------------------------------------------->

	// Header Array Sizes
	private final static byte FileHeaderSize             = (byte)22; // FileHeader is always 22 bytes for first icon in ICO format
	private final static byte InfoHeaderSize             = (byte)40; // InfoHeader is always 40 bytes in size for BMP/ICO/CUR format

	// File Header Fields
	private final static byte TypeByte1                  = (byte)1; // Icons are always Type 1
	private final static byte TypeByte2                  = (byte)0;
	private final static byte ReservedByte1              = (byte)0;
	private final static byte ReservedByte2              = (byte)0;
	private final static byte EntryReservedByte          = (byte)0;
	private final static byte PlanesByte1                = (byte)0;
	private final static byte PlanesByte2                = (byte)0;
	private final static byte BitCountByte1              = (byte)0;
	private final static byte BitCountByte2              = (byte)0;

	// Bitmap Info Header Fields
	private final static byte InfoHeaderSizeByte1        = (byte)(InfoHeaderSize         & 0xFF);
	private final static byte InfoHeaderSizeByte2        = (byte)((InfoHeaderSize >> 8)  & 0xFF);
	private final static byte InfoHeaderSizeByte3        = (byte)((InfoHeaderSize >> 16) & 0xFF);
	private final static byte InfoHeaderSizeByte4        = (byte)((InfoHeaderSize >> 24) & 0xFF);
	private final static byte BitmapPlanesByte1          = (byte)1; // Icons always have 1 bitplane
	private final static byte BitmapPlanesByte2          = (byte)0;
	private final static byte BitmapCompressionByte1     = (byte)0;
	private final static byte BitmapCompressionByte2     = (byte)0;
	private final static byte BitmapCompressionByte3     = (byte)0;
	private final static byte BitmapCompressionByte4     = (byte)0;
	private final static byte BitmapImageSizeByte1       = (byte)0;
	private final static byte BitmapImageSizeByte2       = (byte)0;
	private final static byte BitmapImageSizeByte3       = (byte)0;
	private final static byte BitmapImageSizeByte4       = (byte)0;
	private final static byte BitmapXPixelsPerMByte1     = (byte)0;
	private final static byte BitmapXPixelsPerMByte2     = (byte)0;
	private final static byte BitmapXPixelsPerMByte3     = (byte)0;
	private final static byte BitmapXPixelsPerMByte4     = (byte)0;
	private final static byte BitmapYPixelsPerMByte1     = (byte)0;
	private final static byte BitmapYPixelsPerMByte2     = (byte)0;
	private final static byte BitmapYPixelsPerMByte3     = (byte)0;
	private final static byte BitmapYPixelsPerMByte4     = (byte)0;
	private final static byte BitmapColorsImportantByte1 = (byte)0;
	private final static byte BitmapColorsImportantByte2 = (byte)0;
	private final static byte BitmapColorsImportantByte3 = (byte)0;
	private final static byte BitmapColorsImportantByte4 = (byte)0;

//  Variables ---------------------------------------------------------------------------->

	private static int      IntBitmapWidth;  // Width of image in pixels
	private static int      IntBitmapHeight; // Height of image in pixels
	private static int      IntBitmapSize;     // Size of bitmap data
	private static int      IntBitmapMaskSize; // Size of bitmap transparency mask
	private static int      IntBitmapFileSize; // Size of entire bitmap file, including headers
	private static int      ColorsUsed;
	private static byte     BitDepth;
	private static byte[][] ColorTable;
	private static byte[][] XORBitMap;
	private static byte[][] ANDBitMap;

//  Constructor -------------------------------------------------------------------------->

	public ICOCodec()
	{
	}

//  Encode Method ------------------------------------------------------------------------>

	public static void encode(Image image, File ICOfile) throws FileNotFoundException, IOException, AWTException
	{
		convertToFormat(image);
		writeImageToFile(ICOfile);
	}

	public static void encode(Image image, String ICOfilename) throws FileNotFoundException, IOException, AWTException
	{
		encode(image, new File(ICOfilename));
	}

//  Image Convert Method ---------------------------------------------------------------->

	protected static void convertToFormat(Image image) throws AWTException
	{
	}

//  File Write Method ------------------------------------------------------------------->

	private static void writeImageToFile(File outFile) throws FileNotFoundException, IOException
	{
		IntBitmapSize     = (XORBitMap.length * XORBitMap[0].length);
		IntBitmapMaskSize = (ANDBitMap.length * ANDBitMap[0].length);
		IntBitmapFileSize = InfoHeaderSize + IntBitmapSize + IntBitmapMaskSize;

		// values mapped to bytes
		byte CountByte1 = (byte)1; // number of icons in file, currently set to 1 until it works better
		byte CountByte2 = (byte)0;
			// entry data
			byte WidthByte           = (byte)(IntBitmapWidth & 0xFF);
			byte HeightByte          = (byte)(IntBitmapHeight & 0xFF);
			byte ColorCountByte      = (BitDepth == 1 ? (byte)2 : (BitDepth == 4 ? (byte)16 : (byte)0));
			byte SizeInBytesByte1    = (byte)(IntBitmapFileSize & 0xFF);
			byte SizeInBytesByte2    = (BitDepth == 1 ? (byte)1 : (BitDepth == 4 ? (byte)2 : (byte)8)); // some weirdness that makes it work
			byte SizeInBytesByte3    = (byte)0;
			byte SizeInBytesByte4    = (byte)0;
			byte FileOffsetByte1     = (byte)(FileHeaderSize & 0xFF); // hardcoded for first entry, which for now is the only entry
			byte FileOffsetByte2     = (byte)0;
			byte FileOffsetByte3     = (byte)0;
			byte FileOffsetByte4     = (byte)0;
			// icon data
			byte IconWidthByte1      = (byte)(IntBitmapWidth         & 0xFF);
			byte IconWidthByte2      = (byte)((IntBitmapWidth >> 8)  & 0xFF);
			byte IconWidthByte3      = (byte)((IntBitmapWidth >> 16) & 0xFF);
			byte IconWidthByte4      = (byte)((IntBitmapWidth >> 24) & 0xFF);
			byte IconHeightByte1     = (byte)((IntBitmapHeight + IntBitmapHeight)         & 0Xff);
			byte IconHeightByte2     = (byte)(((IntBitmapHeight + IntBitmapHeight) >> 8)  & 0Xff);
			byte IconHeightByte3     = (byte)(((IntBitmapHeight + IntBitmapHeight) >> 16) & 0Xff);
			byte IconHeightByte4     = (byte)(((IntBitmapHeight + IntBitmapHeight) >> 24) & 0Xff);
			byte IconBitCountByte1   = (byte)(BitDepth        & 0xFF);
			byte IconBitCountByte2   = (byte)((BitDepth >> 8) & 0xFF);
			byte IconColorsUsedByte1 = (byte)(ColorsUsed         & 0xFF);
			byte IconColorsUsedByte2 = (byte)((ColorsUsed >> 8)  & 0xFF);
			byte IconColorsUsedByte3 = (byte)((ColorsUsed >> 16) & 0xFF);
			byte IconColorsUsedByte4 = (byte)((ColorsUsed >> 24) & 0xFF);

		byte[] IconFileHeader = new byte[] {
			ReservedByte1, ReservedByte2,
			TypeByte1, TypeByte2,
			CountByte1, CountByte2
		};

		byte[] IconEntryHeader = new byte[] {
			WidthByte,
			HeightByte,
			ColorCountByte,
			EntryReservedByte,
			PlanesByte1,      PlanesByte2,
			BitCountByte1,    BitCountByte2,
			SizeInBytesByte1, SizeInBytesByte2, SizeInBytesByte3, SizeInBytesByte4,
			FileOffsetByte1,  FileOffsetByte2,  FileOffsetByte3,  FileOffsetByte4
		};

		byte[] IconBodyHeader = new byte[] {
			InfoHeaderSizeByte1,        InfoHeaderSizeByte2,        InfoHeaderSizeByte3,        InfoHeaderSizeByte4,
			IconWidthByte1,             IconWidthByte2,             IconWidthByte3,             IconWidthByte4,
			IconHeightByte1,            IconHeightByte2,            IconHeightByte3,            IconHeightByte4,
			BitmapPlanesByte1,          BitmapPlanesByte2,
			IconBitCountByte1,          IconBitCountByte2,
			BitmapCompressionByte1,     BitmapCompressionByte2,     BitmapCompressionByte3,     BitmapCompressionByte4,
			BitmapImageSizeByte1,       BitmapImageSizeByte2,       BitmapImageSizeByte3,       BitmapImageSizeByte4,
			BitmapXPixelsPerMByte1,     BitmapXPixelsPerMByte2,     BitmapXPixelsPerMByte3,     BitmapXPixelsPerMByte4,
			BitmapYPixelsPerMByte1,     BitmapYPixelsPerMByte2,     BitmapYPixelsPerMByte3,     BitmapYPixelsPerMByte4,
			IconColorsUsedByte1,        IconColorsUsedByte2,        IconColorsUsedByte3,        IconColorsUsedByte4,
			BitmapColorsImportantByte1, BitmapColorsImportantByte2, BitmapColorsImportantByte3, BitmapColorsImportantByte4
		};

		FileOutputStream fos = new FileOutputStream(outFile);
		fos.write(IconFileHeader);
		fos.write(IconEntryHeader);
		fos.write(IconBodyHeader);
		if(ColorTable != null)
		{
			for(int i = 0; i < ColorTable.length; i++)
			{
				fos.write(ColorTable[i]);
			}
		}
		for(int i = 0; i < XORBitMap.length; i++)
		{
			fos.write(XORBitMap[i]);
		}
		for(int i = 0; i < ANDBitMap.length; i++)
		{
			fos.write(ANDBitMap[i]);
		}
		fos.flush();
		fos.close();
	}

//  Transformation Method --------------------------------------------------------------->

	public static void transformBMP(File BMPfile, File ICOfile) throws FileNotFoundException, IOException, AWTException
	{
		obtainBMP(BMPfile);
		writeImageToFile(ICOfile);
	}

	public static void transformBMP(String BMPfilename, String ICOfilename) throws FileNotFoundException, IOException, AWTException
	{
		transformBMP(new File(BMPfilename), new File(ICOfilename));
	}

//  Internal Class Methods --------------------------------------------------------------->

	private static void obtainBMP(File inFile) throws FileNotFoundException, IOException, AWTException
	{
		byte[] byteArray = new byte[(int)(inFile.length())];

		// Read in BMP
		FileInputStream fis = new FileInputStream(inFile);
		fis.read(byteArray);
		fis.close();

		// Bitmap Header
		if(byteArray.length < 2)
		{
			throw new AWTException("File (" + inFile.getName() + ") too small to be a valid bitmap");
		}
		String signature = "" + (char)(byteArray[0]) + (char)(byteArray[1]);
		if(!signature.equals("BM"))
		{
			throw new AWTException("File (" + inFile.getName() + ") does not have a bitmap signature");
		}
		IntBitmapWidth  = (int)(byteArray[18] + byteArray[19] + byteArray[20] + byteArray[21]);
		IntBitmapHeight = (int)(byteArray[22] + byteArray[23] + byteArray[24] + byteArray[25]);
		if(!(IntBitmapWidth == 16 && IntBitmapHeight == 16) && !(IntBitmapWidth == 32 && IntBitmapHeight == 32) && !(IntBitmapWidth == 64 && IntBitmapHeight == 64))
		{
			throw new AWTException("Bitmap must be one of the following dimensions: 16x16, 32x32, 64x64");
		}
		BitDepth = (byte)(byteArray[28] + byteArray[29]); /* 1 = monochrome palette. NumColors = 1 , 4 = 4bit palletized. NumColors = 16 , 8 = 8bit palletized. NumColors = 256 , 16 = 16bit RGB. NumColors = 65536 , 24 = 24bit RGB. NumColors = 16M */
		int Compression = (byteArray[30] + byteArray[31] + byteArray[32] + byteArray[33]); /* 0 = BI_RGB   no compression , 1 = BI_RLE8 8bit RLE encoding , 2 = BI_RLE4 4bit RLE encoding */
		if(Compression != 0)
		{
			throw new AWTException("ICOCodec does not support compressed bitmaps");
		}
		ColorsUsed = (byteArray[46] + byteArray[47] + byteArray[48] + byteArray[49]);
		if(BitDepth == 8) // Wrap 8-bit color count properly from byte limitations
		{
			if(ColorsUsed == 1) { ColorsUsed = 256; }
			if(ColorsUsed < 1)  { ColorsUsed = 256 + ColorsUsed; }
		}
		// Color Table
		int byteCounter = 54;
		if(BitDepth <= 8)
		{
			int numberOfColors = (2 << (BitDepth - 1));
			if((ColorsUsed < numberOfColors) && (ColorsUsed > 0)) { numberOfColors = ColorsUsed; }
			ColorTable = new byte[numberOfColors][4];
			for(int i = 0; i < numberOfColors; i++)
			{
				ColorTable[i] = new byte[] { byteArray[byteCounter], byteArray[byteCounter+1], byteArray[byteCounter+2], byteArray[byteCounter+3] };
				byteCounter = byteCounter + 4;
			}
		}
		// Raster Data
		int imagWidth = IntBitmapWidth / (BitDepth == 1 ? 8 : (BitDepth == 4 ? 2 : 1));
		int maskWidth = IntBitmapWidth / 8;
		XORBitMap = new byte[IntBitmapHeight][imagWidth];
		ANDBitMap = new byte[IntBitmapHeight][maskWidth];
		for(int row = 0; row < IntBitmapHeight; row++)
		{
			for(int col = 0; col < imagWidth; col++)
			{
				XORBitMap[row][col] = byteArray[byteCounter];
				byteCounter++;
			}
			for(int scol = 0; scol < maskWidth; scol++)
			{
				ANDBitMap[row][scol] = (byte)0;
			}
		}
	}

//  Main Method -------------------------------------------------------------------------->

	/* Main method
	 * Allows for command line transformation of BMP files to ICO files */
	public static void main(String[] args)
	{
		if(args.length > 1)
		{
			try
			{
				ICOCodec.transformBMP(args[0], args[1]);
			}
			catch(Exception fnfe)
			{
				fnfe.printStackTrace(System.out);
			}
		}
		else
		{
			System.out.println("USAGE : java com.hexidec.codec.ICOCodec BMPfile ICOfile");
		}
	}

}