/*
GNU General Public License

WBMPCodec
Copyright (C) 2001  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.codec;

import java.awt.AWTException;
import java.awt.Image;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.awt.image.PixelGrabber;

/** WBMPCodec
  *
  * @author Howard Kistler
  * @version 0.1
  *
  * VERSION HISTORY
  * 0.1 (01/24/2001) - initial creation (01/24/2001)
  */

public class WBMPCodec
{

//  Constants ---------------------------------------------------------------------------->

	// WAP Bitmap Dimension Constants
	public static final int MAX_WBMP_WIDTH   = 150;  // the maximum width supported by the WBMP standard
	public static final int MAX_WBMP_HEIGHT  = 150;  // the maximum height supported by the WBMP standard
	public static final int OPT_WBMP_WIDTH   = 96;   // the optimal maximum width to support the greatest range of WAP devices
	public static final int OPT_WBMP_HEIGHT  = 65;   // the optimal maximum height to support the greatest range of WAP devices
	public static final int MAX_BYTES        = 1461; // a WBMP file should not be larger than this

	// Render Mode Settings
	public static final int RENDER_ANY = 0; // render the bitmap regardless of original's size
	public static final int RENDER_OPT = 1; // render the bitmap after scaling to the optimal size
	public static final int RENDER_MAX = 2; // render the bitmap after scaling to the maximum size

	// Constant WBMP Fields
	private final static byte ReservedByte1 = (byte)0;
	private final static byte ReservedByte2 = (byte)0;

//  Variables ---------------------------------------------------------------------------->

	private static int IntBitmapWidth;    // Width of image in pixels
	private static int IntBitmapHeight;   // Height of image in pixels
	private static int IntBitmapSize;     // Size of bitmap data
	private static int IntBitmapFileSize; // Size of entire bitmap file, including headers
	private static int[] PixelMap;        // Pixel array of image data

	private static int renderMode;	

//  Constructor -------------------------------------------------------------------------->

	public WBMPCodec(int mode)
	{
		renderMode = mode;
	}

	public WBMPCodec()
	{
		renderMode = RENDER_OPT;
	}

//  Encode Method ------------------------------------------------------------------------>

	public static void encode(Image image, File WBMPfile) throws FileNotFoundException, IOException, AWTException
	{
		convertToFormat(image);
		writeImageToFile(WBMPfile);
	}

	public static void encode(Image image, String WBMPfilename) throws FileNotFoundException, IOException, AWTException
	{
		encode(image, new File(WBMPfilename));
	}

//  Image Convert Method ---------------------------------------------------------------->

	public static void convertToFormat(Image image) throws AWTException
	{
		IntBitmapWidth  = image.getWidth(null);
		IntBitmapHeight = image.getHeight(null);

		if(renderMode == RENDER_OPT)
		{
			if((IntBitmapWidth > OPT_WBMP_WIDTH) || (IntBitmapHeight > OPT_WBMP_HEIGHT))
			{
				float ratioW = (IntBitmapWidth / OPT_WBMP_WIDTH);
				float ratioH = (IntBitmapHeight / OPT_WBMP_HEIGHT);
				if(ratioW >= ratioH)
				{
					image = image.getScaledInstance(OPT_WBMP_WIDTH, -1, Image.SCALE_AREA_AVERAGING);
				}
				else
				{
					image = image.getScaledInstance(-1, OPT_WBMP_HEIGHT, Image.SCALE_AREA_AVERAGING);
				}
				IntBitmapWidth  = image.getWidth(null);
				IntBitmapHeight = image.getHeight(null);
			}
		}
		else if(renderMode == RENDER_MAX)
		{
			if((IntBitmapWidth > MAX_WBMP_WIDTH) || (IntBitmapHeight > MAX_WBMP_HEIGHT))
			{
				float ratioW = (IntBitmapWidth / MAX_WBMP_WIDTH);
				float ratioH = (IntBitmapHeight / MAX_WBMP_HEIGHT);
				if(ratioW >= ratioH)
				{
					image = image.getScaledInstance(MAX_WBMP_WIDTH, -1, Image.SCALE_AREA_AVERAGING);
				}
				else
				{
					image = image.getScaledInstance(-1, MAX_WBMP_HEIGHT, Image.SCALE_AREA_AVERAGING);
				}
				IntBitmapWidth  = image.getWidth(null);
				IntBitmapHeight = image.getHeight(null);
			}
		}

		PixelMap = new int[IntBitmapWidth * IntBitmapHeight];

		PixelGrabber pg = new PixelGrabber(image, 0, 0, IntBitmapWidth, IntBitmapHeight, PixelMap, 0, IntBitmapWidth);
		try
		{
			pg.grabPixels();
		}
		catch(InterruptedException ie)
		{
			ie.printStackTrace(System.out);
		}

		int BitPadding = (4 - ((IntBitmapWidth * 3) % 4));
		if(BitPadding == 4) { BitPadding = 0; }
		BitPadding = BitPadding * IntBitmapHeight;

		IntBitmapSize = ((IntBitmapWidth * IntBitmapHeight) * 3) + BitPadding;
	}

//  File Write Method ------------------------------------------------------------------->

	protected static void writeImageToFile(File BMPfile) throws IOException
	{
		FileOutputStream fos = new FileOutputStream(BMPfile);

		fos.write(ReservedByte1);
		fos.write(ReservedByte2);

		byte BitmapWidthByte1      = (byte)(IntBitmapWidth          & 0xFF);
		byte BitmapWidthByte2      = (byte)((IntBitmapWidth >> 8)   & 0xFF);
		byte BitmapWidthByte3      = (byte)((IntBitmapWidth >> 16)  & 0xFF);
		byte BitmapWidthByte4      = (byte)((IntBitmapWidth >> 24)  & 0xFF);
		byte[] BitmapWidthBytes;
		if(BitmapWidthByte4 != 0)
		{
			BitmapWidthBytes = new byte[] {
				BitmapWidthByte1, BitmapWidthByte2, BitmapWidthByte3, BitmapWidthByte4
			};
		}
		else if(BitmapWidthByte3 != 0)
		{
			BitmapWidthBytes = new byte[] {
				BitmapWidthByte1, BitmapWidthByte2, BitmapWidthByte3
			};
		}
		else if(BitmapWidthByte2 != 0)
		{
			BitmapWidthBytes = new byte[] {
				BitmapWidthByte1, BitmapWidthByte2
			};
		}
		else
		{
			BitmapWidthBytes = new byte[] {
				BitmapWidthByte1
			};
		}
		fos.write(BitmapWidthBytes);

		byte BitmapHeightByte1     = (byte)(IntBitmapHeight         & 0xFF);
		byte BitmapHeightByte2     = (byte)((IntBitmapHeight >> 8)  & 0xFF);
		byte BitmapHeightByte3     = (byte)((IntBitmapHeight >> 16) & 0xFF);
		byte BitmapHeightByte4     = (byte)((IntBitmapHeight >> 24) & 0xFF);
		byte[] BitmapHeightBytes;
		if(BitmapHeightByte4 != 0)
		{
			BitmapHeightBytes = new byte[] {
				BitmapHeightByte1, BitmapHeightByte2, BitmapHeightByte3, BitmapHeightByte4
			};
		}
		else if(BitmapHeightByte3 != 0)
		{
			BitmapHeightBytes = new byte[] {
				BitmapHeightByte1, BitmapHeightByte2, BitmapHeightByte3
			};
		}
		else if(BitmapHeightByte2 != 0)
		{
			BitmapHeightBytes = new byte[] {
				BitmapHeightByte1, BitmapHeightByte2
			};
		}
		else
		{
			BitmapHeightBytes = new byte[] {
				BitmapHeightByte1
			};
		}
		fos.write(BitmapHeightBytes);

		IntBitmapFileSize = 2 + BitmapWidthBytes.length + BitmapHeightBytes.length + IntBitmapSize;
		if(IntBitmapFileSize > MAX_BYTES)
		{
			System.out.println("WBMP file will be larger than recommended maximum (size=" + IntBitmapFileSize + ")");
		}

		// convert image to 8 pixels-per-byte monochrome image
		// pad the columns out to blocks of 8
		int IntBitmapWidthPadded = (int)(Math.ceil(IntBitmapWidth / 8)) * 8;
		// read by row
		for(int h = 0; h < IntBitmapHeight; h++)
		{
			// start a byte block of 8 bits
			byte block = 0;
			int blockCount = 0;
			int getVal = 0;
			// then by column, padding to blocks of 8
			for(int w = 0; w < IntBitmapWidthPadded; w++)
			{
				if(w < IntBitmapWidth)
				{
					getVal = 0; // replace with function that reads source image and converts to black or white pixel
				}
				else
				{
					getVal = 1; // pad with white (should be a configurable option which to pad with)
				}
				if(blockCount > 7)
				{
					fos.write(block);
					blockCount = 0;
					block = (byte)(getVal);
				}
				else
				{
					block += (byte)(getVal >> 2^blockCount);
					blockCount++;
				}
			}
		}

/*
		int bodySize = IntBitmapWidth * IntBitmapHeight;
		int bitPad = 4 - ((IntBitmapWidth * 3) % 4);
		if(bitPad == 4) { bitPad = 0; }
		int countRow = 1;
		int indexRow = bodySize - IntBitmapWidth;
		int indexLastRow = indexRow;
		int vRed;
		int vGreen;
		int vBlue;
		int vAvg;

		for(int i = 0; i < bodySize; i++)
		{
			int colorValue = PixelMap[indexRow];
			vRed   = colorValue         & 0xFF;
			vGreen = (colorValue >> 8)  & 0xFF;
			vBlue  = (colorValue >> 16) & 0xFF;
			vAvg   = (vRed + vGreen + vBlue) / 3;
			fos.write(rgbArray);
			if(countRow == IntBitmapWidth)
			{
				// pad row to 4 bits requirement
				for(int p = 0; p < bitPad; p++)
				{
					fos.write(0x00);
				}
				countRow = 1;
				indexRow = indexLastRow - IntBitmapWidth;
				indexLastRow = indexRow;
			}
			else
			{
				countRow++;
			}
			indexRow++;
		}
*/

		fos.flush();
		fos.close();
	}

}
